import json
import random
from decimal import Decimal


def deprecate(message):
    import warnings

    warnings.warn(message, DeprecationWarning, stacklevel=3)


class TolerantJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        import uuid

        if isinstance(obj, uuid.UUID):
            return str(obj)
        if isinstance(obj, Decimal):
            return int(obj) if int(obj) == obj else float(obj)
        return json.JSONDecoder.default(self, obj)


def _send_message(*, message, content: dict, close: bool):
    message.reply_channel.send(
        {"text": json.dumps(content, cls=TolerantJSONEncoder), "close": close}
    )


def build_message(
    *,
    handler,
    message_id=None,
    error=None,
    generate_id=False,
    data: dict = None,
    **additional_data
):
    content = dict()
    if message_id:
        content["id"] = message_id
    elif generate_id:
        content["id"] = random.randint(a=1, b=1000)

    if handler:
        content["type"] = handler
    if error:
        content["error"] = error
    if data or additional_data:
        content["data"] = data or {}
        if additional_data:
            deprecate(
                "Passing arbitrary kwargs to `build_message()` is depreated."
                " Pass in a dict `data` instead."
            )
            content["data"].update(additional_data)
    return content


def send_message(
    *,
    message,
    handler,
    message_id=None,
    close=False,
    error=None,
    data: dict = None,
    **additional_data
):
    if additional_data:
        deprecate(
            "Passing arbitrary kwargs to `send_message()` is depreated."
            " Pass in a dict `data` instead."
        )
        if data is None:
            data = {}
        data.update(additional_data)
    content = build_message(
        handler=handler, message_id=message_id, error=error, data=data
    )

    if not content:
        raise Exception("Cannot send an empty message.")
    _send_message(message=message, content=content, close=close)


def send_success(
    *, message, handler, message_id, close=False, data: dict = None, **additional_data
):
    """
    This method directly wraps send_message but checks the existence of id and type.
    """
    if message_id is None or handler is None:
        raise Exception(
            "You have to provide a message ID and handler on success messages."
        )

    if additional_data:
        deprecate(
            "Passing arbitrary kwargs to `send_success()` is depreated."
            " Pass in a dict `data` instead."
        )
        if data is None:
            data = {}
        data.update(additional_data)

    send_message(
        message=message, message_id=message_id, handler=handler, close=close, data=data
    )


def send_error(
    *,
    message,
    handler,
    message_id,
    error,
    close=False,
    data: dict = None,
    **additional_data
):
    """
    This method wraps send_message and makes sure that error is a keyword argument.
    """
    if additional_data:
        deprecate(
            "Passing arbitrary kwargs to `send_error()` is depreated."
            " Pass in a dict `data` instead."
        )
        if data is None:
            data = {}
        data.update(additional_data)

    send_message(
        message=message,
        message_id=message_id,
        handler=handler,
        error=error,
        close=close,
        data=data,
    )



import json


class ReplyChannel:
    def __init__(self):
        self.data = []
        self.closed = False

    def send(self, message_dict):
        if not self.closed:
            self.data.append(json.loads(message_dict.get("text")))
            self.closed = message_dict.get("close", False)


class Message:
    def __init__(self):
        self.reply_channel = ReplyChannel()



import json

from .message import TolerantJSONEncoder, build_message

try:
    from channels import Group
except ImportError:
    Group = None
    print("Channels is not installed, running in test mode!")

try:
    from django.db.transaction import on_commit as on_transaction_commit
except ImportError:

    def on_transaction_commit(func):
        func()


class NotifyBase:
    encoder = TolerantJSONEncoder

    def _get_group_name(self, message_type=""):
        if hasattr(self, "get_group_name"):
            return self.get_group_name(message_type=message_type)
        if hasattr(self, "group_name"):
            return self.group_name.format(pk=self.pk)
        return "{class_name}-{pk}".format(
            class_name=self.__class__.__name__.lower(), pk=self.pk
        )

    def _get_push_data(self, updated_fields=None, message_type=None):
        if hasattr(self, "get_push_notification_data"):
            obj_data = self.get_push_notification_data(
                updated_fields=updated_fields, message_type=message_type
            )
        elif hasattr(self, "serializer_class"):
            obj_data = self.serializer_class(self).data
        else:
            obj_data = {"id": self.pk}
        return obj_data

    def _get_message_type(self, message_type):
        return message_type

    def _send_notify(
        self, message_type, updated_fields=None, data=None, always_send=True
    ):
        group_name = self._get_group_name(message_type=message_type)
        message_data = {
            "group": group_name,
            "type": self._get_message_type(message_type),
            "object": data
            or self._get_push_data(
                updated_fields=updated_fields, message_type=message_type
            ),
        }
        if updated_fields and "updated_fields" not in message_data["object"]:
            message_data["object"]["updated_fields"] = updated_fields

        if message_type == "update" and updated_fields and always_send is False:
            if not set(updated_fields) & set(message_data["object"].keys()):
                return

        payload = {
            "text": json.dumps(
                build_message(
                    generate_id=True, handler="subscription-update", data=message_data
                ),
                cls=self.encoder,
            )
        }

        on_transaction_commit(lambda: Group(group_name).send(payload))


class NotifyOnCreate(NotifyBase):
    def save(self, *args, **kwargs):
        initial_pk = self.pk
        ret = super().save(*args, **kwargs)
        if not initial_pk:
            self._send_notify("create", updated_fields=kwargs.get("update_fields"))
        return ret


class NotifyOnUpdate(NotifyBase):
    rohrpost_always_send = (
        True
    )  # Toggle if even data not reflecting upldate_fields should be sent

    def save(self, *args, **kwargs):
        initial_pk = self.pk
        ret = super().save(*args, **kwargs)
        if initial_pk:
            self._send_notify(
                "update",
                updated_fields=kwargs.get("update_fields"),
                always_send=self.rohrpost_always_send,
            )
        return ret


class NotifyOnDelete(NotifyBase):
    def delete(self, *args, **kwargs):
        self._send_notify("delete")
        return super().delete(*args, **kwargs)


class NotifyOnChange(NotifyOnCreate, NotifyOnUpdate, NotifyOnDelete):
    """
    This model mixin class sends rohrpost messages every time an object is created, updated, or
    deleted, to the corresponding group. The group is named, most important first:
        - As the result of get_group_name() if the object has such a method
        - As the result of group_name.format(pk=object.pk) if the object/class has a group_name
          attribute
        - As the result of '{class_name}-{pk}'.format(class_name=object.__class__.__name__.lower()
          pk=object.pk) otherwise
    The serialized object will be taken either from:
        - The result of get_push_notification_data(updated_fields=None) if such a method exists, else
        - The result of serializer_class(self).data, if such an attribute exists, else
        - {"id": self.pk}
    The serialized object will also contain an 'updated_fields' attribute with a list *if*
    'updated_fields' was not set by the get_push_notification_data() or the serializer_class
    *and* if it was set when calling the save() method.
    get_push_notification_data() receives a list of updated_fields.

    The message will look like this:
    {
        "id": <some id>,
        "type": "subscription-update",
        "data": {
            "type": <create|update|delete>,
            "group": <group-name>,
            "object": <serialized object>,
        }
    }
    """

    pass



HANDLERS = dict()


def _rohrpost_handler(func, name):
    def ret_func(*args, **kwargs):
        return func(*args, **kwargs)

    def update_registry(key, value):
        if key in HANDLERS:
            raise Exception('Handler for "{}" is already defined.'.format(key))
        HANDLERS[key] = value

    name = name or func.__name__

    if isinstance(name, list):
        for n in name:
            update_registry(n, func)
    else:
        update_registry(name, func)
    return ret_func


def rohrpost_handler(name=""):
    def wrap(f):
        return _rohrpost_handler(f, name)

    return wrap



from rohrpost.message import send_message
from rohrpost.registry import rohrpost_handler


@rohrpost_handler("ping")
def handle_ping(message, request):
    """
    Handles requests of this format ("data" being an optional attribute):
    {
        "id": 123,
        "type": "ping",
        "data": {
            "some": "data"
        },
    }
    Response (again, with "data" optional but the same as in the request):
    {
        "id": 123,
        "type": "pong",
        "data": {
            "some": "data"
        },
    }
    Caution: if the sent data is not a dict, it will be wrapped in an
    additional "data" object in the response.
    """
    data = None
    if "data" in request:
        if isinstance(request["data"], dict):
            data = request["data"]
        else:
            data = {"data": request["data"]}

    send_message(message=message, message_id=request["id"], handler="pong", data=data)



from setuptools import setup

from rohrpost import __version__


def read(filepath):
    with open(filepath, "r", encoding="utf-8") as f:
        return f.read()


setup(
    name="rohrpost",
    version=__version__,
    description="rohrpost WebSocket protocol for ASGI",
    long_description=read("README.rst"),
    url="https://github.com/axsemantics/rohrpost",
    author="Tobias Kunze",
    author_email="tobias.kunze@ax-semantics.com",
    license="MIT",
    python_requires=">=3.4",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
    ],
    packages=["rohrpost"],
)



import json
from functools import partial

from . import handlers  # noqa
from .message import send_error
from .registry import HANDLERS

REQUIRED_FIELDS = ["type", "id"]


try:
    DECODE_ERRORS = (json.JSONDecodeError, TypeError)
except AttributeError:
    # Python 3.4 raises a ValueError instead of json.JSONDecodeError
    DECODE_ERRORS = (ValueError, TypeError)


def handle_rohrpost_message(message):
    """
    Handling of a rohrpost message will validate the required format:
    A valid JSON object including at least an "id" and "type" field.
    It then hands off further handling to the registered handler (if any).
    """
    _send_error = partial(send_error, message=message, message_id=None, handler=None)
    if not message.content["text"]:
        return _send_error(error="Received empty message.")

    try:
        request = json.loads(message.content["text"])
    except DECODE_ERRORS as e:
        return _send_error(
            error="Could not decode JSON message. Error: {}".format(str(e))
        )

    if not isinstance(request, dict):
        return _send_error(error="Expected a JSON object as message.")

    for field in REQUIRED_FIELDS:
        if field not in request:
            return _send_error(error="Missing required field '{}'.".format(field))

    if not request["type"] in HANDLERS:
        return send_error(
            message=message,
            message_id=request["id"],
            handler=request["type"],
            error="Unknown message type '{}'.".format(request["type"]),
        )

    HANDLERS[request["type"]](message, request)
